#ifndef _CAMERA_INL_
#define _CAMERA_INL_

#include "Camera.h"
#include "Matrix.inl"
#include "Point.inl"
#include "Vector.inl"

inline Camera& Camera::operator=(const Camera& c)
{
  fCP = c.fCP;
  fVPN = c.fVPN;
  fCOA = c.fCOA;
  fVUP = c.fVUP;
  fNear = c.fNear;                 //Near clipping plane
  fFar = c.fFar;                  //Far clipping plane
  fHeight = c.fHeight;               //Height of projection plane

  fCameraMatrix = c.fCameraMatrix;
  fPerspectiveMatrix = c.fPerspectiveMatrix;
  fWindowMatrix = c.fWindowMatrix;

  return *this;
}

//Meber access functions
inline Vector Camera::GetCP() const
{
  return fCP;
}

inline Vector Camera::GetVPN() const
{
  return fVPN;
}

inline Vector Camera::GetCOA() const
{
  return fCOA;
}

inline Vector Camera::GetVUP() const
{
  return fVUP;
}

inline double Camera::GetNear() const
{
  return fNear;
}

inline double Camera::GetFar() const
{
  return fFar;
}

inline double Camera::GetHeight() const
{
  return fHeight;
}

inline void Camera::Set(double *coa, double *pos, double *vup, double _height_, double _near_, double _far_)
{
  fCOA = Vector(coa[0], coa[1], coa[2]);
  fCP = Vector(pos[0], pos[1], pos[2]);
  fVUP = Vector(vup[0], vup[1], vup[2]);
  fVPN = Normalize(Vector(coa[0] - pos[0], coa[1] - pos[1], coa[2] - pos[2]));
  fNear = _near_;
  fFar = _far_;
  fHeight = _height_;

  //WindowXform depends on CameraXform, so call CameraXform first
  fCameraMatrix = ViewMatrix(fCOA, fCP, fVUP);
  fPerspectiveMatrix = PerspectiveMatrix(fNear, fFar, fHeight);
  PrepareWindowMatrix ();
}

inline void Camera::PrepareWindowMatrix ()
{
  double midx = 0,
         midy = 0,
         s = 0;
  Point center,
        bottomLeft,
        topRight;

  //Find window center
  s = fNear / Magnitude(fVPN);
  center.Set(fCP[0] + (fVPN[0] * s),  fCP[1] + (fVPN[1] * s), fCP[2] + (fVPN[2] * s));

  //Find window coordinates
  midx = fHeight/2.0; 
  midy = fHeight/2.0;

  bottomLeft.Set(center.GetX() - midx, center.GetY() - midy, center.GetZ());
  topRight.Set(center.GetX() + midx, center.GetY() + midy, center.GetZ());

  //Apply transform to window
  bottomLeft = ApplyCamera (bottomLeft);
  topRight = ApplyCamera (topRight);
  bottomLeft = ApplyPerspective (bottomLeft);
  topRight = ApplyPerspective (topRight);
    
  fWindowMatrix = TranslationMatrix((topRight.GetX() - bottomLeft.GetX())/2.0, (topRight.GetY() - bottomLeft.GetY())/2.0, 0.0) * ScaleMatrix(1.0/(topRight.GetX() - bottomLeft.GetX()), 1.0/(topRight.GetY() - bottomLeft.GetY()), 1.0);
}

inline Point Camera::ApplyCamera (const Point& in) const
{
  Point ret = fCameraMatrix * in;
  return ret;
}
  
inline Point Camera::ApplyPerspective (const Point& in) const
{
  Point ret = fPerspectiveMatrix * in;
  ret.Homogenize ();
  return ret;
}
  
inline Point Camera::ApplyWindow (const Point& in) const
{
  Point ret = fWindowMatrix * in;
  return ret;
}

inline Point ApplyCamera(const Camera& C, const Point& in)
{
  return C.ApplyCamera(in);
}

inline Point ApplyPerspective (const Camera& C, const Point& in)
{
  return C.ApplyPerspective(in);
}

inline Point ApplyWindow (const Camera& C, const Point& in)
{
  return C.ApplyWindow(in);
}

#endif