#ifndef _MATRIX_INL_
#define _MATRIX_INL_

#include "Matrix.h"
#include "Point.inl"
#include "Vector.inl"

inline void Matrix::Invert()
{
  int i, j, k;
  Matrix inverse = IdentityMatrix();

  //Current fMatrix will be set to identity while
  //inverse is set with inverse matrix of current

  for (i = 0; i < 4; i++)
  {
    if (fMatrix[i][i] != 1.0)
    {
      double divby = fMatrix[i][i];
      for (j = 0; j < 4; j++)
      {
        inverse.fMatrix[i][i] /= divby;
        fMatrix[i][j] /= divby;
      }
    }
    for (j = 0; j < 4; j++)
    {
      if (j != i)
      {
        if (fMatrix[j][i] != 0.0)
        {
          double mulby = fMatrix[j][i];
          for (k = 0; k < 4; k++)
          {
            fMatrix[j][k] -= mulby * fMatrix[i][k];
            inverse.fMatrix[j][k] -= mulby * inverse.fMatrix[i][k];
          }
        }
      }
    }
  }

  *this = inverse;
}

inline void Matrix::Transpose()
{
  int i, j;

  for (i = 0; i < 4; i++)
  {
    for (j = 1; j < 4; j++)
    {
      if (i != j)
      {
        double temp = fMatrix[i][j];
        fMatrix[i][j] = fMatrix[j][i];
        fMatrix[j][i] = temp;
      }
    }
  }
}

inline void Matrix::SetIdentity()
{
    int i, j;

    for (i = 0; i < 4; i++)
      for (j = 0; j < 4; j++)
        fMatrix[i][j] = (i == j) ? 1.0 : 0.0;
}

inline Matrix Invert(const Matrix& m)
{
  Matrix invert = m;
  invert.Invert();
  return invert;
}

inline Matrix IdentityMatrix()
{
  Matrix m;
  m.SetIdentity();
  return m;
}

inline Matrix Transpose(const Matrix& m)
{
  Matrix trans = m;
  trans.Transpose();
  return trans;
}

inline Matrix ZeroMatrix()
{
  Matrix M;
  int i, j;

  for (i = 0; i < 4; i++)
    for (j = 0; j < 4; j++)
      M.fMatrix[i][j] = 0.0;
  return M;
}

inline Matrix TranslationMatrix(double x, double y, double z)
{
  Matrix M = IdentityMatrix();
  M.fMatrix[3][0] = x;
  M.fMatrix[3][1] = y;
  M.fMatrix[3][2] = z;
  return M;
}

inline Matrix RotationXMatrix(double angle)
{
  Matrix M = IdentityMatrix();
  double Cosine = cos(angle);
  double Sine = sin(angle);

  M.fMatrix[1][1] = Cosine;
  M.fMatrix[2][1] = -Sine;
  M.fMatrix[1][2] = Sine;
  M.fMatrix[2][2] = Cosine;
  return M;
}

inline Matrix RotationYMatrix(double angle)
{
  Matrix M = IdentityMatrix();
  double Cosine = cos(angle);
  double Sine = sin(angle);

  M.fMatrix[0][0] = Cosine;
  M.fMatrix[2][0] = -Sine;
  M.fMatrix[0][2] = Sine;
  M.fMatrix[2][2] = Cosine;
  return M;
}

inline Matrix RotationZMatrix(double angle)
{
  Matrix M = IdentityMatrix();
  double Cosine = cos(angle);
  double Sine = sin(angle);

  M.fMatrix[0][0] = Cosine;
  M.fMatrix[1][0] = -Sine;
  M.fMatrix[0][1] = Sine;
  M.fMatrix[1][1] = Cosine;
  return M;
}

inline Matrix ScaleMatrix(double x, double y, double z)
{
  Matrix M = IdentityMatrix();
  M.fMatrix[0][0] = x;
  M.fMatrix[1][1] = y;
  M.fMatrix[2][2] = z;
  return M;
}

inline Matrix ViewMatrix(const Vector& LookAt, const Vector& Viewer, const Vector& UpL)
{
  int i;
  Matrix M = IdentityMatrix();
  Matrix T = TranslationMatrix(-Viewer[0], -Viewer[1], -Viewer[2]);
  Vector U, V, N;
  Vector Up = Normalize(UpL);

  N = Normalize(LookAt - Viewer);

  V = Up - (DotProduct(Up, N) * N);
  V = Normalize(V);

  U = CrossProduct(N, V);
  U = Normalize(U);

  for (i = 0; i < 3; i++)
  {
    M.fMatrix[i][0] = U[i]; 
    M.fMatrix[i][1] = V[i]; 
    M.fMatrix[i][2] = N[i]; 
  }

  T *= M;

  return T;
}

inline Matrix operator*(const Matrix& A, const Matrix& B)
{
  Matrix ret;
  int i, j, k;

  for (i = 0; i < 4; i++)
  {
    for (j = 0; j < 4; j++)
    {
      double subt = 0;
      for (k = 0; k < 4; k++)
        subt += A.fMatrix[i][k] * B.fMatrix[k][j];
      ret.fMatrix[i][j] = subt;
    }
  }

  return ret;
}

inline Vector operator*(const Matrix& M, const Vector& v)
{
  Vector ret;
  double denom;

  ret.fZ = v.fX * M.fMatrix[0][0] + 
           v.fY * M.fMatrix[1][0] + 
           v.fZ * M.fMatrix[2][0] +
           M.fMatrix[3][0];

  ret.fY = v.fX * M.fMatrix[0][1] + 
           v.fY * M.fMatrix[1][1] + 
           v.fZ * M.fMatrix[2][1] +
           M.fMatrix[3][1];

  ret.fZ = v.fX * M.fMatrix[0][2] + 
           v.fY * M.fMatrix[1][2] + 
           v.fZ * M.fMatrix[2][2] +
           M.fMatrix[3][2];

  denom = M.fMatrix[0][3] + M.fMatrix[1][3] + M.fMatrix[2][3] + M.fMatrix[3][3];
  if (denom != 1.0)
    ret /= denom;

  return ret;
}

inline Point operator*(const Matrix& M, const Point& p)
{
  Point ret = p; //copy p's contents - color, normal , etc

  ret.fX = p.fX * M.fMatrix[0][0] + 
           p.fY * M.fMatrix[1][0] + 
           p.fZ * M.fMatrix[2][0] +
           p.fH * M.fMatrix[3][0];

  ret.fY = p.fX * M.fMatrix[0][1] + 
           p.fY * M.fMatrix[1][1] + 
           p.fZ * M.fMatrix[2][1] +
           p.fH * M.fMatrix[3][1];

  ret.fZ = p.fX * M.fMatrix[0][2] + 
           p.fY * M.fMatrix[1][2] + 
           p.fZ * M.fMatrix[2][2] +
           p.fH * M.fMatrix[3][2];

  ret.fH = p.fX * M.fMatrix[0][3] + 
           p.fY * M.fMatrix[1][3] + 
           p.fZ * M.fMatrix[2][3] +
           p.fH * M.fMatrix[3][3];

  return ret;
}

inline Matrix& Matrix::operator*=(const Matrix& M)
{
  Matrix ret = *this;
  int i, j, k;

  for (i = 0; i < 4; i++)
  {
    for (j = 0; j < 4; j++)
    {
      double subt = 0.0;
      for (k = 0; k < 4; k++)
      {
        subt += ret.fMatrix[i][k] * M.fMatrix[k][j];
      }
      fMatrix[i][j] = subt;
    }
  }

  return *this;
}

//Do perspective projection with near and far clipping planes and window height
inline Matrix PerspectiveMatrix (double D, double F, double H)
{
  Matrix M = IdentityMatrix();

  M.fMatrix[2][2] = (H * F) / (D * (F - D));
  M.fMatrix[2][3] = H/D;
  M.fMatrix[3][2] = ((-1.0 * H) * F) / (F - D);
  M.fMatrix[3][3] = 0.0;

  return M;
}

//Project with focal length
inline Matrix PerspectiveMatrix (double depth)
{
  Matrix M = IdentityMatrix();

  M.fMatrix[2][3] = 1.0/depth;
  M.fMatrix[3][3] = 0.0;

  return M;
}

inline Matrix ShadowMatrix(double *groundPlane, const Vector& lightPos)
{
  Vector dir;
  Matrix m = IdentityMatrix();

  double dot;

  dot = groundPlane[0] * lightPos[0] +
        groundPlane[1] * lightPos[1] +
        groundPlane[2] * lightPos[2] +
        groundPlane[3] * 0.0;

  m.fMatrix[0][0] = dot - lightPos[0] * groundPlane[0];
  m.fMatrix[1][0] = 0.0 - lightPos[0] * groundPlane[1];
  m.fMatrix[2][0] = 0.0 - lightPos[0] * groundPlane[2];
  m.fMatrix[3][0] = 0.0 - lightPos[0] * groundPlane[3];

  m.fMatrix[0][1] = 0.0 - lightPos[1] * groundPlane[0];
  m.fMatrix[1][1] = dot - lightPos[1] * groundPlane[1];
  m.fMatrix[2][1] = 0.0 - lightPos[1] * groundPlane[2];
  m.fMatrix[3][1] = 0.0 - lightPos[1] * groundPlane[3];

  m.fMatrix[0][2] = 0.0 - lightPos[2] * groundPlane[0];
  m.fMatrix[1][2] = 0.0 - lightPos[2] * groundPlane[1];
  m.fMatrix[2][2] = dot - lightPos[2] * groundPlane[2];
  m.fMatrix[3][2] = 0.0 - lightPos[2] * groundPlane[3];

  m.fMatrix[0][3] = 0.0 - 0.0 * groundPlane[0];
  m.fMatrix[1][3] = 0.0 - 0.0 * groundPlane[1];
  m.fMatrix[2][3] = 0.0 - 0.0 * groundPlane[2];
  m.fMatrix[3][3] = dot - 0.0 * groundPlane[3];

  return m;
}

#endif
