#ifndef _MODEL_INL_
#define _MODEL_INL_

#include <stdio.h>
#include "wx/wxprec.h"
#include "Model.h"
#include "Environment.h"
#include "PointLight.h"
#include "Vertex.h"
#include "Poly.h"

#include "Vertex.inl"
#include "Poly.inl"
#include "Camera.inl"
#include "Matrix.inl"
#include "PointLight.inl"

inline void Model::SetComputeVertexNormals(bool yn)
{
  fComputeNormals = yn;
}

inline bool Model::ShouldComputeVertexNormals() const
{
  return fComputeNormals;
}

inline void Model::MakeShadow(const Vector& lightPos)
{
  int i;
  for (i = 0; i < fNumPolys; i++)
    fPolys[i].MakeShadow(lightPos);
}


inline void Model::ApplyMatrix(const Matrix& M)
{
  int i;
  for (i = 0; i < fNumPolys; i++)
    fPolys[i].ApplyMatrix(M);
}

inline void Model::CameraTransform(const Camera& C)
{
  int i;
  for (i = 0; i < fNumPolys; i++)
    fPolys[i].CameraTransform(C);
}

inline void Model::PerspectiveTransform(const Camera& C)
{
  int i;
  for (i = 0; i < fNumPolys; i++)
    fPolys[i].PerspectiveTransform(C);
}

inline void Model::WindowTransform(const Camera& C)
{
  int i;
  for (i = 0; i < fNumPolys; i++)
    fPolys[i].WindowTransform(C);
}

inline void Model::Cull(const Vector& V)
{
  int i;
  for (i = 0; i < fNumPolys; i++)
    fPolys[i].Cull(V);
}

inline void Model::Cull()
{
  int i;
  for (i = 0; i < fNumPolys; i++)
    fPolys[i].Cull();
}

inline void Model::Render(const Camera& C, wxDC& dc)
{
  int i;

  for (i = 0; i < fNumPolys; i++)
  {
    Poly P = ::CameraTransform(fPolys[i], C);

    Vector CP = C.GetCP();
    Point p(CP[0], CP[1], CP[2]);  
	  p = ApplyCamera(C, p);
		Vector pos(p.GetX(), p.GetY(), p.GetZ());

    P.Cull(pos);
    if (!P.Culled())
    {
	    P.PerspectiveTransform(C);
  	  P.WindowTransform(C);

      P.Render (dc);
    }

    //Do shadows if they're requested
    if (Environment::SHADOWTYPE == Environment::FAKESHADOW)
    {
      int j;
      unsigned store1;
      bool store2;
      for (j = 0; j < Environment::fNumPointLights; j++)
      {
        Poly S = ::MakeShadow(fPolys[i], Environment::fPointLights[j].GetPosition());

        S.CameraTransform(C);

        Vector CP = C.GetCP();
        Point p(CP[0], CP[1], CP[2]);  
	      p = ApplyCamera(C, p);
		    Vector pos(p.GetX(), p.GetY(), p.GetZ());

        S.Cull(pos);
        if (!S.Culled())
        {
	        S.PerspectiveTransform(C);
  	      S.WindowTransform(C);

          S.SetColor(0.0, 0.0, 0.0);

          if (Environment::RENDERTYPE != Environment::WIREFRAME)
          {
            store1 = Environment::RENDERTYPE;
            store2 = Environment::LIGHTSCENE;
            Environment::RENDERTYPE = Environment::FLAT;
            Environment::LIGHTSCENE = false;
            S.Render (dc);
            Environment::LIGHTSCENE = store2;
            Environment::RENDERTYPE = store1;
          }
          else
            S.Render (dc);
        }
      }
    }
  
  }
}

inline void Model::ComputeVertexNormals (const char *file)
{
  Vector normal;
  Vector current;
  int total = 0;
  int i, j;
  FILE *fp = NULL;

  if (file)
  {
    fp = fopen(file, "w");
    fprintf(fp, "%d\n", fNumVerts);
  }

  //Process each vertex once
  for (i = 0; i < fNumVerts; i++)
  {
    //Reset total
    normal[0] = 0.0;
    normal[1] = 0.0;
    normal[2] = 0.0;

    //Check each face for the vertex
    for (j = 0; j < fNumPolys; j++)
    {
      //If the face contains the vertex, add it to the normal average
      if (fPolys[j].ContainsVertex(i))
      {
        //Sum normals
        current = fPolys[j].ComputeSurfaceNormal ();
        normal += current;
      }
    }

    //Calculate average
    normal = Normalize (normal);

    //Now set the normal for the faces that contain this vertex
    for (j = 0; j < fNumPolys; j++)
    {
      fPolys[j].ContainsVertex(i);
        fPolys[j].SetVertexNormal(i, -normal);
    }

    //Write to file if needed
    if (fp)
    {
      fprintf(fp, "%f %f %f\n", normal[0], normal[1], normal[2]);
    }
  }

  if (fp)
    fclose(fp);
}

inline void Model::ComputeVertexColors ()
{
  int i;
  for (i = 0; i < fNumPolys; i++)
  {
    fPolys[i].ComputeVertexColors();
  }
}

inline Model MakeShadow(const Model& s, const Vector& lightPos)
{
  Model ret = s;
  ret.MakeShadow(lightPos);
  return ret;
}

inline Model ApplyMatrix(const Model& s, const Matrix& M)
{
  Model ret = s;
  ret.ApplyMatrix(M);
  return ret;
}

inline Model CameraTransform(const Model& s, const Camera& C)
{
  Model ret = s;
  ret.CameraTransform(C);
  return ret;
}

inline Model PerspectiveTransform(const Model& s, const Camera& C)
{
  Model ret = s;
  ret.PerspectiveTransform(C);
  return ret;
}

inline Model WindowTransform(const Model& s, const Camera& C)
{
  Model ret = s;
  ret.WindowTransform(C);
  return ret;
}

inline Model Cull(const Model& s, const Vector& V)
{
  Model ret = s;
  ret.Cull(V);
  return ret;
}

inline Model Cull(const Model& s)
{
  Model ret = s;
  ret.Cull();
  return ret;
}

inline void Render(Model& s, const Camera& C, wxDC& dc)
{
  s.Render(C, dc);
}

#endif
