#ifndef _POINT_LIGHT_INL_
#define _POINT_LIGHT_INL_

#include "PointLight.h"

#include "RGBColor.inl"
#include "Vector.inl"
#include "Point.inl"

inline PointLight& PointLight::operator=(const PointLight& p)
{
  fIntensity = p.fIntensity;
  fPosition = p.fPosition;
  fDirection = p.fDirection;
  fP = p.fP;
  fC1 = p.fC1;
  fC2 = p.fC2;
  fC3 = p.fC3;
  fAttenuationType = p.fAttenuationType;
  fSpotAngle = p.fSpotAngle;
  fSpotEnabled = p.fSpotEnabled;
  fLightType = p.fLightType;
  memcpy(fFlaps, p.fFlaps, sizeof(fFlaps));
  memcpy(fFlapEnabled, p.fFlapEnabled, sizeof(fFlapEnabled));

  return *this;
}

inline void PointLight::SetLightType(unsigned type)
{
  fLightType = type;
}

inline unsigned PointLight::GetLightType() const
{
  return fLightType;
}

inline void PointLight::EnableFlap(unsigned flap)
{
  fFlapEnabled[flap] = true;
}

inline void PointLight::DisableFlap(unsigned flap)
{
  fFlapEnabled[flap] = false;
}

inline bool PointLight::IsFlapEnabled(unsigned flap) const
{
  return fFlapEnabled[flap];
}

inline bool PointLight::AreFlapsEnabled() const
{
  return (fFlapEnabled[FLAPMINX] || fFlapEnabled[FLAPMAXX] || fFlapEnabled[FLAPMINY] || fFlapEnabled[FLAPMAXY] || fFlapEnabled[FLAPMINZ] || fFlapEnabled[FLAPMAXZ]);
}

inline void PointLight::EnableSpotLight()
{
  fSpotEnabled = true;
}

inline void PointLight::DisableSpotLight()
{
  fSpotEnabled = false;
}

inline bool PointLight::IsSpotLightEnabled() const
{
  return fSpotEnabled;
}

inline void PointLight::SetIntensity(double r, double g, double b)
{
  fIntensity[0] = r;
  fIntensity[1] = g;
  fIntensity[2] = b;
}

inline void PointLight::SetPosition(double r, double g, double b)
{
  fPosition[0] = r;
  fPosition[1] = g;
  fPosition[2] = b;
}

inline void PointLight::SetPosition(const Vector& v)
{
  fPosition = v;
}

inline void PointLight::SetIntensity(const RGBColor& v)
{
  fIntensity = v;
}

inline void PointLight::SetDirection(double x, double y, double z)
{
  fDirection[0] = x;
  fDirection[1] = y;
  fDirection[2] = z;

  fDirection = Normalize(fDirection);
}

inline void PointLight::SetDirection(const Vector& v)
{
  fDirection = Normalize(v);
}

inline void PointLight::SetDirectionExponent(double p)
{
  fP = p;
}

inline void PointLight::SetSpotAngle(double angle)
{
  //Store as radians
  fSpotAngle = angle * DEG_TO_RAD;
}

inline double PointLight::GetSpotAngle() const
{
  return fSpotAngle * RAD_TO_DEG;
}

inline void PointLight::SetFlapValue(unsigned flap, double restriction)
{
  fFlaps[flap] = restriction;
}

inline double PointLight::GetFlapValue(unsigned flap) const
{
  return fFlaps[flap];
}

inline Vector PointLight::GetDirection () const
{
  return fDirection;
}

inline double PointLight::GetDirectionExponent() const
{
  return fP;
}

inline Vector PointLight::GetPosition() const
{
  return fPosition;
}

inline RGBColor PointLight::GetIntensity(const Vector& Dir) const
{
  if (fLightType == DIRECTIONAL)
  {
    double cosg;
    Vector L = Normalize(Dir);
    cosg = DotProduct(-L, fDirection);

    return (fIntensity * pow(cosg, fP));
  }
  else
  {
    return fIntensity;
  }
}

inline void PointLight::SetAttenuationValues(double c1, double c2, double c3)
{
  fC1 = c1;
  fC2 = c2;
  fC3 = c3;
}

inline void PointLight::GetAttenuationValues(double *c1, double *c2, double *c3)
{
  *c1 = fC1;
  *c2 = fC2;
  *c3 = fC3;
}

inline double PointLight::GetAttenuation(const Point &pos) const
{
  if (fAttenuationType == NO_ATTENUATION)
    return 1;
  else
  {
    double a;
    double d = Magnitude(Vector(fPosition[0] - pos.GetX(), fPosition[1] - pos.GetY(), fPosition[2] - pos.GetZ()));
    
    if (fAttenuationType == LINEAR_ATTENUATION)
      a = 1.0/(d * d);
    else //QUADRATIC_ATTENUATION
      a = 1.0/(fC1 + (fC2 * d) + (fC3 * d * d));

    return min(a, 1.0);
  }
}

inline bool PointLight::IsInFlaps(const Point &pos) const
{
  //No flaps if its not Warn lighting
  if (fLightType == NONDIRECTIONAL)
    return true;
    
  //XMin
  if (fFlapEnabled[FLAPMINX])
  {
    if (pos.GetX() < fFlaps[FLAPMINX])
        return false;
  }
  //XMax
  if (fFlapEnabled[FLAPMAXX])
  {
    if (pos.GetX() > fFlaps[FLAPMAXX])
        return false;
  }

  //YMin
  if (fFlapEnabled[FLAPMINY])
  {
    if (pos.GetY() < fFlaps[FLAPMINY])
        return false;
  }
  //YMax
  if (fFlapEnabled[FLAPMAXY])
  {
    if (pos.GetY() > fFlaps[FLAPMAXY])
        return false;
  }

  //ZMin
  if (fFlapEnabled[FLAPMINZ])
  {
    if (pos.GetZ() < fFlaps[FLAPMINZ])
        return false;
  }
  //ZMax
  if (fFlapEnabled[FLAPMAXZ])
  {
    if (pos.GetZ() > fFlaps[FLAPMAXZ])
        return false;
  }

  return true;
}

inline bool PointLight::IsInSpot(const Vector &Dir) const
{
  double cosg, 
         cosp = cos(fSpotAngle);
  Vector L = Normalize(Dir);    //Just in case

  //If it's not a spot light, true
  if (!fSpotEnabled || fLightType == NONDIRECTIONAL)
    return true;
  
  cosg = DotProduct(-L, fDirection);
  if (fabs(cosg) > fabs(cosp))
  {
    return true;
  }

  return false;
}

#endif