#ifndef _POLY_INL_
#define _POLY_INL_

#include "wx/wxprec.h"
#include "Poly.h"
#include "Environment.h"
#include "ScanFill.h"

#include "Camera.h"
#include "Matrix.h"

#include "Camera.inl"
#include "Matrix.inl"
#include "RGBColor.inl"
#include "SurfaceInfo.inl"
#include "Vertex.inl"
#include "ScanFill.inl"
#include "FakeShadow.inl"

//Method to do Cartesian to pixel conversions
inline int XToDevice(double x)
{
  return (int)(x * ((double) Environment::WIDTH));
}

inline int YToDeviceRaw (double y)
{
  return (int)(y * ((double) Environment::HEIGHT));
}

inline int YToDevice (double y)
{
  return (Environment::HEIGHT - YToDeviceRaw (y));
}

//Pixel to Cartesian
inline double XToScreen (int x)
{
  return ((double)x / ((double)Environment::WIDTH));
}

inline double YToScreenRaw (double y)
{
  return ((double)y / ((double)Environment::HEIGHT));
}

inline double YToScreen (double y)
{
  return ((double)(Environment::HEIGHT - y) / ((double)Environment::HEIGHT));
}

inline void Poly::Set(double *x, double *y, double *z, int nverts)
{
  fCulled = false;

  if (fVertices)
    delete [] fVertices;
  fVertices = new Vertex[nverts];
  fNVerts = nverts;
    
  for (int i = 0; i < nverts; i++)
  {
    fVertices[i].Set(x[i], y[i], z[i]);
  }
}

inline void Poly::Set(const Vertex *pts, int nverts)
{
  fCulled = false;

  if (fVertices)
    delete [] fVertices;
  fVertices = new Vertex[nverts];
  fNVerts = nverts;
    
  for (int i = 0; i < nverts; i++)
  {
    fVertices[i] = pts[i];
  }
}

inline void Poly::SetColor(double r, double g, double b)
{
  fColor[0] = r;
  fColor[1] = g;
  fColor[2] = b;
  fSurfaceInfo.SetObjectDiffuseColor(r, g, b);
}

inline void Poly::SetColor(const RGBColor& rgb)
{
  fColor = rgb;
  fSurfaceInfo.SetObjectDiffuseColor(rgb);
}

inline void Poly::SetSurfaceInfo(const SurfaceInfo& info)
{
  fSurfaceInfo = info;
}

inline RGBColor Poly::GetColor() const
{
  return fColor;
}

inline void Poly::SetTextureMap(TextureMap *map)
{
  fTextureMap = map;
}

inline void Poly::SetFakeShadow(const FakeShadow& s)
{
  fFakeShadow = s;
}

inline SurfaceInfo Poly::GetSurfaceInfo() const
{
  return fSurfaceInfo;
}

inline Poly MakeShadow(const Poly& s, const Vector& lightPos)
{
  Poly apply = s;
  apply.MakeShadow(lightPos);
  return apply;
}

inline Poly ApplyMatrix(const Poly& s, const Matrix& M)
{
  Poly apply = s;
  apply.ApplyMatrix(M);
  return apply;
}

inline Poly CameraTransform(const Poly& s, const Camera& C)
{
  Poly apply = s;
  apply.CameraTransform(C);
  return apply;
}

inline Poly PerspectiveTransform(const Poly& s, const Camera& C)
{
  Poly apply = s;
  apply.PerspectiveTransform(C);
  return apply;
}

inline Poly WindowTransform(const Poly& s, const Camera& C)
{
  Poly apply = s;
  apply.WindowTransform(C);
  return apply;
}

inline void Poly::Render(wxDC& dc)
{
  //Copy points to arrays
  int i;

  if (Environment::RENDERTYPE == Environment::WIREFRAME)
  {
    wxPoint *xyPoints = new wxPoint[fNVerts];

    //Set color
    wxColour col(255 * fColor[0], 255 * fColor[1], 255 * fColor[2]);
    wxPen colpen(col, 1, wxSOLID);
    dc.SetPen(colpen);

    for (i = 0; i < fNVerts; i++)
    {
      //Convert to normalized screen coordinates
      xyPoints[i].x = XToDevice(fVertices[i].GetX());
      xyPoints[i].y = YToDevice(fVertices[i].GetY());
    }

    dc.DrawPolygon (fNVerts, xyPoints);
    delete [] xyPoints;
  }
  else
  {
    ScanFill polyScan;
    ScanPoint *xyzPoints = new ScanPoint[fNVerts];

    //If it's phong each point will have it's color computed by the scan converter
    if (Environment::RENDERTYPE != Environment::PHONG)
      ComputeVertexColors();

    for (i = 0; i < fNVerts; i++)
    {
      //Convert to normalized screen coordinates
      xyzPoints[i].x = XToDevice(fVertices[i].GetX());
      xyzPoints[i].y = YToDevice(fVertices[i].GetY());
      xyzPoints[i].z = fVertices[i].GetZ();
      xyzPoints[i].i = fVertices[i].GetColor();
      xyzPoints[i].n = fVertices[i].GetNormal();
      xyzPoints[i].wc = fVertices[i].GetWC();
    }

    polyScan.SetSurfaceInfo(&fSurfaceInfo);
    polyScan.SetTextureMap(fTextureMap);

    polyScan.Scan (dc, fNVerts, xyzPoints);
    delete [] xyzPoints;
  }

}

//Apply a matrix transformation to the polygon
inline void Poly::MakeShadow(const Vector& lightPos)
{
  int i;

  //Set fake shadow for current light
  fFakeShadow.ComputeShadowMatrix(lightPos);

  for (i = 0; i < fNVerts; i++)
  {
    //Apply transformation to each point
    fVertices[i] = ::ApplyShadow(fVertices[i], fFakeShadow);
  }
}

//Apply a matrix transformation to the polygon
inline void Poly::ApplyMatrix(const Matrix& M)
{
  int i;

  for (i = 0; i < fNVerts; i++)
  {
    //Apply transformation to each point
    fVertices[i] = ::ApplyMatrix(fVertices[i], M);
  }
}

//Apply transformation from world to eye
inline void Poly::CameraTransform(const Camera& C)
{
  int i;

  for (i = 0; i < fNVerts; i++)
  {
    //Apply transformation to each point
    fVertices[i] = ::CameraTransform(fVertices[i], C);
  }
}

//Apply perspective projection to polygon
inline void Poly::PerspectiveTransform(const Camera& C)
{
  int i;

  for (i = 0; i < fNVerts; i++)
  {
    //Apply transformation to each point
    fVertices[i] = ::PerspectiveTransform(fVertices[i], C);
  }
}

//Rotate window to viewport
inline void Poly::WindowTransform (const Camera& C)
{
  int i;

  for (i = 0; i < fNVerts; i++)
  {
    //Apply transformation to each point
    fVertices[i] = ::WindowTransform(fVertices[i], C);
  }
}

//Cull in eye space
inline bool Poly::Cull (const Vector& V)
{
  double dot;
  	
  //CCW vector
  Vector V1(fVertices[1].GetX() - fVertices[0].GetX(), fVertices[1].GetY() - fVertices[0].GetY(), fVertices[1].GetZ() - fVertices[0].GetZ());
  //CW vector
  Vector V2(fVertices[2].GetX() - fVertices[0].GetX(), fVertices[2].GetY() - fVertices[0].GetY(), fVertices[2].GetZ() - fVertices[0].GetZ());
  Vector NP = CrossProduct(V1, V2);
  NP = Normalize (NP);

  Vector view(V[0] - fVertices[1].GetX(), V[1] - fVertices[1].GetY(), V[2] - fVertices[1].GetZ());
  view = Normalize(view);

  dot = DotProduct(NP, view);
  fCulled = (dot > 0) ? false : true;
      
  return fCulled;   
}

//Cull in Screen space
inline bool Poly::Cull ()
{
  //CCW vector
  Vector V1(fVertices[1].GetX() - fVertices[0].GetX(), fVertices[1].GetY() - fVertices[0].GetY(), fVertices[1].GetZ() - fVertices[0].GetZ());
  //CW vector
  Vector V2(fVertices[2].GetX() - fVertices[0].GetX(), fVertices[2].GetY() - fVertices[0].GetY(), fVertices[2].GetZ() - fVertices[0].GetZ());
  Vector NP = CrossProduct(V1, V2);
  NP = Normalize (NP);
    
  fCulled = (NP[2] <= 0) ? false : true;

  return fCulled;
}

inline Poly Cull(const Poly& s, const Vector& V)
{
  Poly ret = s;
  ret.Cull(V);
  return ret;
}

inline Poly Cull(const Poly& s)
{
  Poly ret = s;
  ret.Cull();
  return ret;
}

inline void Render(Poly& s, wxDC& dc)
{
  s.Render(dc);
}

inline Poly& Poly::operator=(const Poly& p)
{
  fCulled = p.fCulled;
  fNVerts = p.fNVerts;
  fColor = p.fColor;
  fSurfaceInfo = p.fSurfaceInfo;
  fTextureMap = p.fTextureMap;
  fFakeShadow = p.fFakeShadow;
  fVertices = new Vertex[p.fNVerts];
    
  for (int i = 0; i < p.fNVerts; i++)
  {
    fVertices[i] = p.fVertices[i];
  }

  return *this;
}

inline bool Poly::ContainsVertex(int id) const
{
  int i;
  for (i = 0; i < fNVerts; i++)
  {
    if (fVertices[i].GetID() == id)
      return true;
  }

  return false;
}

inline Vector Poly::ComputeSurfaceNormal()
{
  //CCW vector
  Vector V1(fVertices[1].GetX() - fVertices[0].GetX(), fVertices[1].GetY() - fVertices[0].GetY(), fVertices[1].GetZ() - fVertices[0].GetZ());
  //CW vector
  Vector V2(fVertices[2].GetX() - fVertices[0].GetX(), fVertices[2].GetY() - fVertices[0].GetY(), fVertices[2].GetZ() - fVertices[0].GetZ());
  Vector NP = CrossProduct (V1, V2);

  return Normalize(NP);
}

inline void Poly::SetVertexNormal(int id, const Vector& normal)
{
  int i;
  for (i = 0; i < fNVerts; i++)
  {
    if (fVertices[i].GetID() == id)
      fVertices[i].SetNormal(normal);
  }
}

inline void Poly::ComputeVertexColors()
{
  int i;

  for (i = 0; i < fNVerts; i++)
  {
    //Compute with world coordinate for lighting
    fSurfaceInfo.ComputeIntensity (fVertices[i].GetWC(), fVertices[i].GetNormal());
    fVertices[i].SetColor(fSurfaceInfo.GetIntensity());
  }
}

#endif