#include <stdio.h>
#include <string.h>
#include "Environment.h"
#include "Point.h"
#include "Poly.h"
#include "Model.h"
#include "PointLight.h"
#include "AmbientLight.h"

#include "Environment.inl"
#include "Camera.inl"
#include "Point.inl"
#include "Poly.inl"
#include "Model.inl"
#include "PointLight.inl"
#include "AmbientLight.inl"
#include "Vector.inl"

#ifdef WIN32
#define strncasecmp _strnicmp
#endif

int Environment::WIDTH = 0;
int Environment::HEIGHT = 0;

bool Environment::LIGHTSCENE = true;

const unsigned Environment::WIREFRAME = 1;
const unsigned Environment::FLAT = 2;
const unsigned Environment::GOURARD = 3;
const unsigned Environment::PHONG = 4;
unsigned Environment::RENDERTYPE = Environment::WIREFRAME;

const unsigned Environment::NOSHADOW = 1;
const unsigned Environment::FAKESHADOW = 2;
const unsigned Environment::ZBUFFERSHADOW = 3;
unsigned Environment::SHADOWTYPE = Environment::NOSHADOW;

//Default to white ambient light
AmbientLight Environment::fAmbientLight(1.0);
int Environment::fNumPointLights = 0;
PointLight *Environment::fPointLights = NULL;

Camera Environment::fCamera;
DepthBuffer Environment::fDepthBuffer;

int Environment::fCurrentShadowBuffer = -1;
int Environment::fNumShadowBuffers = 0;
DepthBuffer *Environment::fShadowBuffers = NULL;


Environment::~Environment()
{
  if (fModels)
    delete [] fModels;
}

void Environment::ReadEnvironmentFile(const char *file)
{
  FILE *env;
  char line[1024];
  char *token;
  double col[3];

  env = fopen(file, "r");
  fgets(line, sizeof(line), env);

  //Decide what type 
  if(strncasecmp(line, "WIREFRAME", strlen("WIREFRAME")) == 0)
    RENDERTYPE = WIREFRAME;
  else if(strncasecmp(line, "FLAT", strlen("FLAT")) == 0)
    RENDERTYPE = FLAT;
  else if(strncasecmp(line, "GOURARD", strlen("GOURARD")) == 0)
    RENDERTYPE = GOURARD;
  else if(strncasecmp(line, "PHONG", strlen("PHONG")) == 0)
    RENDERTYPE = PHONG;

  fgets(line, sizeof(line), env);

  //Decide what type 
  if(strncasecmp(line, "NOSHADOW", strlen("NOSHADOW")) == 0)
    SHADOWTYPE = NOSHADOW;
  else if(strncasecmp(line, "FAKESHADOW", strlen("FAKESHADOW")) == 0)
    SHADOWTYPE = FAKESHADOW;
  else if(strncasecmp(line, "ZBUFFERSHADOW", strlen("ZBUFFERSHADOW")) == 0)
    SHADOWTYPE = ZBUFFERSHADOW;

  fgets(line, sizeof(line), env);
  token = strtok(line, " \t");
  col[0] = atof(token);
  token = strtok(NULL, " \t");
  col[1] = atof(token);
  token = strtok(NULL, " \t");
  col[2] = atof(token);

  SetBackgroundColor(col[0], col[1], col[2]);

  fclose (env);
}

void Environment::ReadObjectFiles(const char *file)
{
  FILE *objects;
  char line[1024];
  char *objfile;
  char *colfile;
  char *surfile;
  char *texturefile;
  int i;

  objects = fopen(file, "r");
  fgets(line, sizeof(line), objects);
  fNumModels = atoi(line);

  fModels = new Model[fNumModels];
  for (i = 0; i < fNumModels; i++)
  {
    char *newline;
    fgets(line, sizeof(line), objects);
    newline = strchr(line, '\n');
    if (newline)
      *newline = '\0';

    objfile = strtok(line, " \t");
    colfile = strtok(NULL, " \t");
    surfile = strtok(NULL, " \t");
    texturefile = strtok(NULL, " \t");

    fModels[i].ReadObjectFile(objfile);
    //Read color if there's a file
    if (colfile)
      fModels[i].ReadColorFile(colfile);

    //Read surface properties, if there's a file
    if (surfile)
      fModels[i].ReadSurfaceInfoFile(surfile);

    if (texturefile)
      fModels[i].ReadTextureFile(texturefile);

    //Load the normals if they're needed for the lighting model
    if (RENDERTYPE != WIREFRAME)
    {
      char name[1024];
      char *tok = strtok(objfile, ".");
      strcpy(name, tok);
      strcat(name, ".n");
      FILE *fp = fopen(name, "r");

      if (fp != NULL)
      {
        fclose(fp);
        fModels[i].ReadNormalFile(name);
        fModels[i].SetComputeVertexNormals(false);
      }
      else
      {
        fModels[i].ComputeVertexNormals(name);
        fModels[i].SetComputeVertexNormals(true);
      }
    }
  }

  fclose(objects);
}

void Environment::ReadCameraFile(const char *file)
{
  double coa[3];
  double pos[3];
  double vup[3];
  double bottomleft[2];
  double topright[2];
  double nearplane;
  double farplane;
  FILE *fp;
  char line[1024];
  char *token;

  fp = fopen(file, "r");

  //Read input from file
  fgets(line, sizeof(line), fp);
  token = strtok(line, " \t");
  coa[0] = atof(token);
  token = strtok(NULL, " \t");
  coa[1] = atof(token);
  token = strtok(NULL, " \t");
  coa[2] = atof(token);

  fgets(line, sizeof(line), fp);
  token = strtok(line, " \t");
  pos[0] = atof(token);
  token = strtok(NULL, " \t");
  pos[1] = atof(token);
  token = strtok(NULL, " \t");
  pos[2] = atof(token);

  fgets(line, sizeof(line), fp);
  token = strtok(line, " \t");
  vup[0] = atof(token);
  token = strtok(NULL, " \t");
  vup[1] = atof(token);
  token = strtok(NULL, " \t");
  vup[2] = atof(token);

  fgets(line, sizeof(line), fp);
  token = strtok(line, " \t");
  bottomleft[0] = atof(token);
  token = strtok(NULL, " \t");
  bottomleft[1] = atof(token);

  token = strtok(NULL, " \t");
  topright[0] = atof(token);
  token = strtok(NULL, " \t");
  topright[1] = atof(token);

  fgets(line, sizeof(line), fp);
  token = strtok(line, " \t");
  nearplane = atof(token);
  token = strtok(NULL, " \t");
  farplane = atof(token);

  //Setup camera
  fCamera.Set(coa, pos, vup, topright[1] - bottomleft[1], nearplane, farplane);

  fclose(fp);
}

void Environment::ReadAmbientLightFile(const char *file)
{
  double col[3];
  FILE *fp;
  char line[1024];
  char *token;

  fp = fopen(file, "r");

  //Read input from file
  fgets(line, sizeof(line), fp);
  token = strtok(line, " \t");
  col[0] = atof(token);
  token = strtok(NULL, " \t");
  col[1] = atof(token);
  token = strtok(NULL, " \t");
  col[2] = atof(token);

  fAmbientLight.SetIntensity(col[0], col[1], col[2]);

  fclose(fp);
}

void Environment::ReadPointLightFile(const char *file)
{
  int i;
  int total;
  int flag;
  double restriction;
  double col[3];
  FILE *fp;
  char line[1024];
  char *token;

  fp = fopen(file, "r");

  //Read input from file
  fgets(line, sizeof(line), fp);
  total = atoi (line);

  //If shadow zbuffer, create the buffers
  if (SHADOWTYPE == ZBUFFERSHADOW)
  {
    fNumShadowBuffers = total;
    fShadowBuffers = new DepthBuffer[total];

    for ( i = 0; i < total; i++)
    {
      fShadowBuffers[i].SetSize(WIDTH, HEIGHT);
    }
  }

  fNumPointLights = total;
  fPointLights = new PointLight[fNumPointLights];

  for (i = 0; i < fNumPointLights; i++)
  {
    fgets(line, sizeof(line), fp);
    token = strtok(line, " \t");
    col[0] = atof(token);
    token = strtok(NULL, " \t");
    col[1] = atof(token);
    token = strtok(NULL, " \t");
    col[2] = atof(token);
    fPointLights[i].SetIntensity(col[0], col[1], col[2]);

    fgets(line, sizeof(line), fp);
    token = strtok(line, " \t");
    col[0] = atof(token);
    token = strtok(NULL, " \t");
    col[1] = atof(token);
    token = strtok(NULL, " \t");
    col[2] = atof(token);
    fPointLights[i].SetPosition(col[0], col[1], col[2]);

    fgets(line, sizeof(line), fp);
    token = strtok(line, " \t");
    col[0] = atof(token);
    token = strtok(NULL, " \t");
    col[1] = atof(token);
    token = strtok(NULL, " \t");
    col[2] = atof(token);
    fPointLights[i].SetAttenuationValues(col[0], col[1], col[2]);

    fgets(line, sizeof(line), fp);
    token = strtok(line, " \t");
    col[0] = atof(token);
    token = strtok(NULL, " \t");
    col[1] = atof(token);
    token = strtok(NULL, " \t");
    col[2] = atof(token);
    fPointLights[i].SetDirection(col[0], col[1], col[2]);

    fgets(line, sizeof(line), fp);
    token = strtok(line, " \t");
    col[0] = atoi(token);
    if (col[0])
      fPointLights[i].SetLightType(PointLight::DIRECTIONAL);
    else
      fPointLights[i].SetLightType(PointLight::NONDIRECTIONAL);

    fgets(line, sizeof(line), fp);
    token = strtok(line, " \t");
    col[0] = atof(token);
    fPointLights[i].SetDirectionExponent(col[0]);

    //Read spot light angle
    fgets(line, sizeof(line), fp);
    token = strtok(line, " \t");
    flag = atoi(token);
    if (flag)
    {
      token = strtok(NULL, " \t");
      restriction = atof(token);
      fPointLights[i].EnableSpotLight();
      fPointLights[i].SetSpotAngle(restriction);
    }
    else
    {
      fPointLights[i].DisableSpotLight();
    }

    //Read xmin
    fgets(line, sizeof(line), fp);
    token = strtok(line, " \t");
    flag = atoi(token);
    if (flag)
    {
      token = strtok(NULL, " \t");
      restriction = atof(token);
      fPointLights[i].EnableFlap(0);
      fPointLights[i].SetFlapValue(0, restriction);
    }
    else
    {
      fPointLights[i].DisableFlap(0);
    }

    //Read xmax
    fgets(line, sizeof(line), fp);
    token = strtok(line, " \t");
    flag = atoi(token);
    if (flag)
    {
      token = strtok(NULL, " \t");
      restriction = atof(token);
      fPointLights[i].EnableFlap(1);
      fPointLights[i].SetFlapValue(1, restriction);
    }
    else
    {
      fPointLights[i].DisableFlap(1);
    }

    //Read ymin
    fgets(line, sizeof(line), fp);
    token = strtok(line, " \t");
    flag = atoi(token);
    if (flag)
    {
      token = strtok(NULL, " \t");
      restriction = atof(token);
      fPointLights[i].EnableFlap(2);
      fPointLights[i].SetFlapValue(2, restriction);
    }
    else
    {
      fPointLights[i].DisableFlap(2);
    }

    //Read ymax
    fgets(line, sizeof(line), fp);
    token = strtok(line, " \t");
    flag = atoi(token);
    if (flag)
    {
      token = strtok(NULL, " \t");
      restriction = atof(token);
      fPointLights[i].EnableFlap(3);
      fPointLights[i].SetFlapValue(3, restriction);
    }
    else
    {
      fPointLights[i].DisableFlap(3);
    }

    //Read zmin
    fgets(line, sizeof(line), fp);
    token = strtok(line, " \t");
    flag = atoi(token);
    if (flag)
    {
      token = strtok(NULL, " \t");
      restriction = atof(token);
      fPointLights[i].EnableFlap(4);
      fPointLights[i].SetFlapValue(4, restriction);
    }
    else
    {
      fPointLights[i].DisableFlap(4);
    }

    //Read zmax
    fgets(line, sizeof(line), fp);
    token = strtok(line, " \t");
    flag = atoi(token);
    if (flag)
    {
      token = strtok(NULL, " \t");
      restriction = atof(token);
      fPointLights[i].EnableFlap(5);
      fPointLights[i].SetFlapValue(5, restriction);
    }
    else
    {
      fPointLights[i].DisableFlap(5);
    }

  }

  fclose(fp);
}

void Environment::ReadShadowFile(const char *file)
{
  int i, j;
  int total;
  FILE *objects;
  char line[1024];
  char *newline;

  objects = fopen(file, "r");
  fgets(line, sizeof(line), objects);
  total = atoi (line);

  for (i = 0; i < total; i++)
  {
    //Get the shadowfile name
    fgets(line, sizeof(line), objects);
    newline = strrchr(line, '\n');
    if (newline)
      *newline = '\0';

    //Color every nth polygon
    for (j = i; j < fNumModels; j += total)
      fModels[j].ReadShadowFile(line);
  }

  fclose (objects);
}
